import os
import platform
import re
import subprocess
import sys
from distutils.version import LooseVersion

from setuptools import Extension, setup
from setuptools.command.build_ext import build_ext

PACKAGE_NAME = "qflow"
BACKEND_NAME = f"_{PACKAGE_NAME}_backend"


class CMakeExtension(Extension):
    def __init__(self, name, sourcedir=""):
        Extension.__init__(self, name, sources=[])
        self.sourcedir = os.path.abspath(sourcedir)


class CMakeBuild(build_ext):
    def run(self):
        try:
            out = subprocess.check_output(["cmake", "--version"])
        except OSError:
            raise RuntimeError(
                "CMake must be installed to build the following extensions: "
                + ", ".join(e.name for e in self.extensions)
            )

        if platform.system() == "Windows":
            cmake_version = LooseVersion(
                re.search(r"version\s*([\d.]+)", out.decode()).group(1)
            )
            if cmake_version < "3.1.0":
                raise RuntimeError("CMake >= 3.1.0 is required on Windows")

        for ext in self.extensions:
            self.build_extension(ext)

    def build_extension(self, ext):
        extdir = os.path.abspath(os.path.dirname(self.get_ext_fullpath(ext.name)))
        cmake_args = [
            "-DCMAKE_LIBRARY_OUTPUT_DIRECTORY=" + extdir,
            "-DPYTHON_EXECUTABLE=" + sys.executable,
        ]

        env = os.environ.copy()

        cfg = env.get("CMAKE_BUILD_TYPE", "Debug" if self.debug else "Release")
        build_args = ["--config", cfg]

        if platform.system() == "Windows":
            cmake_args += [
                "-DCMAKE_LIBRARY_OUTPUT_DIRECTORY_{}={}".format(cfg.upper(), extdir)
            ]
            if sys.maxsize > 2 ** 32:
                cmake_args += ["-A", "x64"]
            build_args += ["--", "/m"]
        else:
            cmake_args += ["-DCMAKE_BUILD_TYPE=" + cfg]
            build_args += ["--", "-j5", BACKEND_NAME]
            if cfg.lower() == "coverage":
                build_args += ["gtest"]

        env["CXXFLAGS"] = '{} -DVERSION_INFO=\\"{}\\"'.format(
            env.get("CXXFLAGS", ""), self.distribution.get_version()
        )
        self.build_temp = os.path.join(os.path.dirname(__file__), "build-py")
        if not os.path.exists(self.build_temp):
            os.makedirs(self.build_temp)
        subprocess.check_call(
            " ".join(["cmake", ext.sourcedir] + cmake_args),
            cwd=self.build_temp,
            env=env,
            shell=True,
        )
        subprocess.check_call(
            " ".join(["cmake", "--build", "."] + build_args),
            cwd=self.build_temp,
            env=env,
            shell=True,
        )


def build(setup_kwargs):
    setup_kwargs.update(
        dict(
            ext_modules=[CMakeExtension(BACKEND_NAME)],
            cmdclass=dict(build_ext=CMakeBuild),
            zip_safe=False,
        )
    )



import numpy as np
import scipy.stats as st
import math


def compute_statistics_for_series(x, method="plain", **method_kwargs):
    """Return statistics related to the series x."""
    supported_methods = {"plain": lambda x, **kwargs: x, "blocking": blocking}

    if method not in supported_methods:
        raise NotImplementedError(
            f"The method {method} is not supported. Available options: {supported_methods}"
        )

    method = supported_methods[method]

    x = np.asarray(x)
    x = method(x, **method_kwargs)

    return {
        "mean": np.mean(x),
        "max": np.max(x),
        "min": np.min(x),
        "var": np.var(x),
        "std": np.std(x),
        "sem": st.sem(x),
        "CI": st.t.interval(0.95, len(x) - 1, loc=np.mean(x), scale=st.sem(x)),
    }


def statistics_to_tex(
    all_stats, labels, filename=None, quantity_name="$\\langle E_L\\rangle$"
):
    """
    Produce LaTeX table from statistics produced from ``compute_statistics_for_series``.
    """

    digs = max([abs(int(math.floor(math.log10(stats["sem"])))) for stats in all_stats])

    tex = f"""\\begin{{tabular}}{{lS[table-format=1.%d]*2{{S[table-format=1.%d]}}*2{{S[table-format=1.1]}}}}
\\toprule
\\addlinespace
& {{%s}} & {{CI$^{{95}}_-$}} & {{CI$^{{95}}_+$}} & {{Std}} & {{Var}} \\\\
\\addlinespace
\\midrule
\\addlinespace
\\addlinespace
    """ % (
        digs + 2,
        digs,
        quantity_name,
    )

    for stats, label in zip(all_stats, labels):
        stats = stats.copy()
        stats["ci-"], stats["ci+"] = stats.pop("CI")

        tex += label + " & "

        significant_digits = abs(int(math.floor(math.log10(stats["sem"]))))
        tex += "{0:.{2}f}({1:.0f}) & ".format(
            stats["mean"], stats["sem"] * 10 ** significant_digits, significant_digits
        )
        tex += "{0:.{1}f} & ".format(stats["ci-"], significant_digits)
        tex += "{0:.{1}f} & ".format(stats["ci+"], significant_digits)
        tex += "\\num{{{0:.1e}}} & ".format(stats["std"])
        tex += "\\num{{{0:.1e}}}".format(stats["var"])

        tex += "\\\\\n"

    tex += "\\addlinespace\\addlinespace\\bottomrule\n\\end{tabular}"

    if filename is not None:
        with open(filename, "w") as f:
            f.write(tex)

    return tex


def blocking(x):
    """
    Return an improved estimate of the standard error
    of the mean of the given time series, accounting for
    covariant samples.
    Code by Marius Jonsson.
    Adapted by Bendik Samseth
    """
    # preliminaries
    n = len(x)
    d = int(np.log2(n))
    x_blocks = []
    s = np.empty(d)
    gamma = np.empty(d)
    mu = np.mean(x)

    assert 2 ** d == len(x)  # Size of array must be a power of two.

    # estimate the auto-covariance and variances
    # for each blocking transformation
    for i in range(0, d):
        n = len(x)
        # estimate autocovariance of x
        gamma[i] = np.sum((x[0 : (n - 1)] - mu) * (x[1:n] - mu)) / n
        # Estimate variance
        s[i] = np.var(x)
        # perform blocking transformation
        x = 0.5 * (x[0::2] + x[1::2])
        x_blocks.append(x)

    # generate the test observator M_k from the theorem
    M = (np.cumsum(((gamma / s) ** 2 * 2 ** np.arange(1, d + 1)[::-1])[::-1]))[::-1]

    # we need a list of magic numbers
    q = np.array(
        [
            6.634_897,
            9.210_340,
            11.344_867,
            13.276_704,
            15.086_272,
            16.811_894,
            18.475_307,
            20.090_235,
            21.665_994,
            23.209_251,
            24.724_970,
            26.216_967,
            27.688_250,
            29.141_238,
            30.577_914,
            31.999_927,
            33.408_664,
            34.805_306,
            36.190_869,
            37.566_235,
            38.932_173,
            40.289_360,
            41.638_398,
            42.979_820,
            44.314_105,
            45.641_683,
            46.962_942,
            48.278_236,
            49.587_884,
            50.892_181,
        ]
    )

    # use magic to determine when we should have stopped blocking
    for k in range(0, min(d, len(q))):
        if M[k] < q[k]:
            break

    if k >= d - 1 and s[k] != 0:
        raise RuntimeWarning(
            "Blocking warning: Blocked until stopped and var is not zero. You probably need more data."
        )

    # Return the new data points, corresponding to the best block size.
    return x_blocks[k]



from typing import Sequence, Callable
from abc import ABC, abstractmethod

from .hamiltonians import Hamiltonian
from .wavefunctions import Wavefunction
from .samplers import Sampler
from .optimizers import SgdOptimizer
from .mpi import mpiprint


class Callback(ABC, list):
    @abstractmethod
    def __call__(self, iter_count, psi, H, sampler, optimizer):
        pass


class EnergyCallback(Callback):
    def __init__(self, samples, verbose=False):
        self.samples, self.verbose = samples, verbose

    def __call__(self, iter_count, psi, H, sampler, optimizer):
        self.append(H.local_energy(sampler, psi, self.samples))
        if self.verbose:
            mpiprint(
                f"EnergyCallback(iter={iter_count}): {self.__getitem__(-1)}", flush=True
            )


class SymmetryCallback(Callback):
    def __init__(self, samples, permutations=50, verbose=False):
        self.samples, self.permutations, self.verbose = samples, permutations, verbose

    def __call__(self, iter_count, psi, H, sampler, optimizer):
        self.append(psi.symmetry_metric(sampler, self.samples, self.permutations))
        if self.verbose:
            mpiprint(
                f"SymmetryCallback(iter={iter_count}): {self.__getitem__(-1)}",
                flush=True,
            )


class ParameterCallback(Callback):
    def __call__(self, iter_count, psi, *args):
        self.append([p for p in psi.parameters])


def train(
    psi: Wavefunction,
    H: Hamiltonian,
    sampler: Sampler,
    iters: int,
    samples: int,
    gamma: float,
    optimizer: SgdOptimizer,
    verbose: bool = False,
    call_backs: Sequence[Callable] = tuple(),
    call_back_resolution: int = 100,
):
    """Train a wavefunction according to settings, with call back functions."""
    completed_iterations = 0
    for iteration in range(call_back_resolution):
        for cb in call_backs:
            cb(completed_iterations, psi, H, sampler, optimizer)

        iterations_now = iters // call_back_resolution + int(
            iteration < iters % call_back_resolution
        )
        H.optimize_wavefunction(
            psi, sampler, iterations_now, samples, optimizer, gamma, verbose
        )
        completed_iterations += iterations_now

    for cb in call_backs:
        cb(completed_iterations, psi, H, sampler, optimizer)



from mpi4py import MPI
from pprint import pprint

def master_rank():
    return MPI.COMM_WORLD.rank == 0

def mpiprint(*args, pretty=False, **kwargs):
    if master_rank():
        if pretty:
            pprint(*args, *kwargs)
        else:
            print(*args, **kwargs)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['qflow',
 'qflow.hamiltonians',
 'qflow.optimizers',
 'qflow.samplers',
 'qflow.wavefunctions',
 'qflow.wavefunctions.nn',
 'qflow.wavefunctions.nn.activations',
 'qflow.wavefunctions.nn.layers']

package_data = \
{'': ['*']}

install_requires = \
['mpi4py>=3.0.3,<4.0.0']

setup_kwargs = {
    'name': 'qflow-vmc',
    'version': '0.2.0',
    'description': 'Quantum Variational Monte Carlo',
    'long_description': '[![Build Status](https://travis-ci.org/bsamseth/qflow.svg?branch=master)](https://travis-ci.org/bsamseth/qflow)\n[![codecov](https://codecov.io/gh/bsamseth/qflow/branch/master/graph/badge.svg)](https://codecov.io/gh/bsamseth/qflow)\n[![Language grade: Python](https://img.shields.io/lgtm/grade/python/g/bsamseth/qflow.svg?logo=lgtm&logoWidth=18)](https://lgtm.com/projects/g/bsamseth/qflow/context:python)\n[![Language grade: C/C++](https://img.shields.io/lgtm/grade/cpp/g/bsamseth/qflow.svg?logo=lgtm&logoWidth=18)](https://lgtm.com/projects/g/bsamseth/qflow/context:cpp)\n[![Total alerts](https://img.shields.io/lgtm/alerts/g/bsamseth/qflow.svg?logo=lgtm&logoWidth=18)](https://lgtm.com/projects/g/bsamseth/qflow/alerts/)\n[![Lines of Code](https://tokei.rs/b1/github/bsamseth/qflow)](https://tokei.rs/b1/github/bsamseth/qflow)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/ambv/black)\n\n<p align="center">\n  <img src="docs/logo.png" height="200px">\n</p>\n\n# QFLOW - Quantum Variational Monte Carlo with Neural Networks\n\nSee the [documentation pages](https://bsamseth.github.io/qflow) for details regarding the `qflow` package.\n\nFor the accompanying master\'s thesis, see [writing/Thesis.pdf](https://github.com/bsamseth/qflow/raw/master/writing/Thesis.pdf)\n',
    'author': 'Bendik Samseth',
    'author_email': 'b.samseth@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://bsamseth.github.io/qflow',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}
from build import *
build(setup_kwargs)

setup(**setup_kwargs)



import mpi4py  # Explicitly load mpi4py, which initializes the MPI environment properly.
from _qflow_backend import *
from _qflow_backend import (
    _init_mpi,
    _start_distance_tracking,
    _stop_distance_tracking,
    _set_simulation_box_size,
    _disable_simulation_box,
    _get_simulation_box_size,
)

__all__ = [
    "wavefunctions",
    "hamiltonians",
    "samplers",
    "optimizers",
    "statistics",
    "training",
]

_init_mpi()  # Initializing from Python is not necessary for MPI it self,
             # but rather to seed the random number generators differently
             # for each process.


class SimulationBox(object):
    def __init__(self, L):
        self.L = L

    def __enter__(self):
        self.old_L = _get_simulation_box_size()
        _set_simulation_box_size(self.L)

    def __exit__(self, exc_type, exc_value, traceback):
        if self.old_L < 0:
            _disable_simulation_box()
        else:
            _set_simulation_box_size(self.old_L)


class DistanceCache(object):
    def __init__(self, system, pbc_size=None):
        self.system = system
        self.pbc_size = pbc_size

    def __enter__(self):
        if self.pbc_size is not None:
            _start_distance_tracking(self.system, self.pbc_size)
        else:
            _start_distance_tracking(self.system)

    def __exit__(self, exc_type, exc_value, traceback):
        _stop_distance_tracking()
